# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for the logical behaviour of the tac negotiation skill."""
import copy
from pathlib import Path
from unittest.mock import patch

from aea.decision_maker.gop import GoalPursuitReadiness, OwnershipState, Preferences
from aea.helpers.preference_representations.base import (
    linear_utility,
    logarithmic_utility,
)
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.tac_control.helpers import (
    determine_scaling_factor,
    generate_utility_params,
)
from packages.fetchai.skills.tac_negotiation.dialogues import FipaDialogue
from packages.fetchai.skills.tac_negotiation.strategy import Strategy

from tests.conftest import ROOT_DIR


class TestLogical(BaseSkillTestCase):
    """Logical Tests for tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        tac_dm_context_kwargs = {
            "goal_pursuit_readiness": GoalPursuitReadiness(),
            "ownership_state": OwnershipState(),
            "preferences": Preferences(),
        }
        super().setup(dm_context_kwargs=tac_dm_context_kwargs)
        cls.register_as = "both"
        cls.search_for = "both"
        cls.is_contract_tx = False
        cls.ledger_id = "some_ledger_id"
        cls.location = {"longitude": 0.1270, "latitude": 51.5194}
        cls.search_radius = 5.0
        cls.service_key = "tac_service"

        cls.strategy = Strategy(
            register_as=cls.register_as,
            search_for=cls.search_for,
            is_contract_tx=cls.is_contract_tx,
            ledger_id=cls.ledger_id,
            location=cls.location,
            service_key=cls.service_key,
            search_radius=cls.search_radius,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )

        cls.sender = "some_sender_address"
        cls.counterparty = "some_counterparty_address"

        cls.mocked_currency_id = "12"
        cls.mocked_currency_amount = 2000000
        cls.mocked_amount_by_currency_id = {
            cls.mocked_currency_id: cls.mocked_currency_amount
        }
        cls.mocked_good_ids = ["13", "14", "15", "16", "17", "18", "19", "20", "21"]
        cls.mocked_good_quantities = [5, 7, 4, 3, 5, 4, 3, 5, 6]
        cls.mocked_quantities_by_good_id = dict(
            zip(cls.mocked_good_ids, cls.mocked_good_quantities)
        )
        cls.mocked_ownership_state = (
            cls._skill.skill_context.decision_maker_handler_context.ownership_state
        )
        cls.mocked_ownership_state.set(
            cls.mocked_amount_by_currency_id, cls.mocked_quantities_by_good_id
        )

        cls.exchange_params_by_currency_id = {cls.mocked_currency_id: 1.0}
        cls.utility_params_by_good_id = generate_utility_params(
            [cls._skill.skill_context.agent_address],
            cls.mocked_good_ids,
            determine_scaling_factor(cls.mocked_currency_amount),
        )[cls._skill.skill_context.agent_address]
        cls.mocked_preferences = (
            cls._skill.skill_context.decision_maker_handler_context.preferences
        )
        cls.mocked_preferences.set(
            exchange_params_by_currency_id=cls.exchange_params_by_currency_id,
            utility_params_by_good_id=cls.utility_params_by_good_id,
        )

    @staticmethod
    def _calculate_score(preferences, ownership_state):
        """Calculate the score given a preferences and an ownership_state object."""
        goods_score = logarithmic_utility(
            preferences.utility_params_by_good_id,
            ownership_state.quantities_by_good_id,
        )
        money_score = linear_utility(
            preferences.exchange_params_by_currency_id,
            ownership_state.amount_by_currency_id,
        )
        return goods_score + money_score

    def test_generated_proposals_increase_score_seller(self):
        """Test whether the proposals generated by _generate_candidate_proposals method of the Strategy class actually increases agent's score where role is seller."""
        # setup
        is_searching_for_sellers = True

        # operation
        with patch.object(
            self.skill.skill_context.transactions,
            "ownership_state_after_locks",
            return_value=self.mocked_ownership_state,
        ) as mock_ownership:
            actual_proposals = self.strategy._generate_candidate_proposals(
                is_searching_for_sellers
            )

        # after
        mock_ownership.assert_any_call(is_seller=is_searching_for_sellers)

        current_score = self._calculate_score(
            self.mocked_preferences, self.mocked_ownership_state
        )
        for proposal in actual_proposals:
            # applying proposal on a new ownership_state
            terms = self.strategy.terms_from_proposal(
                proposal, self.sender, self.counterparty, FipaDialogue.Role.SELLER
            )
            new_ownership_state = copy.copy(self.mocked_ownership_state)
            new_ownership_state.apply_delta(
                terms.amount_by_currency_id, terms.quantities_by_good_id
            )

            # new score
            new_score = self._calculate_score(
                self.mocked_preferences, new_ownership_state
            )

            assert new_score >= current_score

    def test_generated_proposals_increase_score_buyer(self):
        """Test whether the proposals generated by _generate_candidate_proposals method of the Strategy class actually increases agent's score  where role is buyer."""
        # setup
        is_searching_for_sellers = False

        # operation
        with patch.object(
            self.skill.skill_context.transactions,
            "ownership_state_after_locks",
            return_value=self.mocked_ownership_state,
        ) as mock_ownership:
            actual_proposals = self.strategy._generate_candidate_proposals(
                is_searching_for_sellers
            )

        # after
        mock_ownership.assert_any_call(is_seller=is_searching_for_sellers)

        current_score = self._calculate_score(
            self.mocked_preferences, self.mocked_ownership_state
        )
        for proposal in actual_proposals:
            # applying proposal on a new ownership_state
            terms = self.strategy.terms_from_proposal(
                proposal, self.sender, self.counterparty, FipaDialogue.Role.BUYER
            )
            new_ownership_state = copy.copy(self.mocked_ownership_state)
            new_ownership_state.apply_delta(
                terms.amount_by_currency_id, terms.quantities_by_good_id
            )

            # new score
            new_score = self._calculate_score(
                self.mocked_preferences, new_ownership_state
            )

            assert new_score >= current_score



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the RegistrationDB class of the confirmation aw2 skill."""
import datetime
import json
import logging
from pathlib import Path
from unittest.mock import Mock, patch

from packages.fetchai.skills.confirmation_aw2.registration_db import RegistrationDB

from tests.conftest import ROOT_DIR
from tests.test_packages.test_skills.test_confirmation_aw2.intermediate_class import (
    ConfirmationAW2TestCase,
)


class TestStrategy(ConfirmationAW2TestCase):
    """Test RegistrationDB of confirmation aw2."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "confirmation_aw2")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.custom_path = None
        cls.db = RegistrationDB(
            custom_path=cls.custom_path,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )
        cls.address = "some_address"
        cls.logger = cls._skill.skill_context.logger

        cls.timestamp = datetime.datetime.now()
        cls.data = {"some_key_1": "some_value_1", "some_key_2": "some_value_2"}

        cls.developer_handle = "developer_handle"

        cls.first_trade = "2020-12-22 18:30:00.000000"
        cls.second_trade = "second_trade"
        cls.first_info = "first_info"
        cls.second_info = "second_info"

    def test__initialise_backend(self):
        """Test the _initialise_backend method of the RegistrationDB class."""
        # operation
        with patch.object(self.db, "_execute_single_sql") as mock_exe:
            self.db._initialise_backend()

        # after
        mock_exe.assert_any_call(
            "CREATE TABLE IF NOT EXISTS registered_table (address TEXT, ethereum_address TEXT, "
            "ethereum_signature TEXT, fetchai_signature TEXT, "
            "developer_handle TEXT, tweet TEXT)"
        )
        mock_exe.assert_any_call(
            "CREATE TABLE IF NOT EXISTS trade_table (address TEXT PRIMARY KEY, first_trade timestamp, "
            "second_trade timestamp, first_info TEXT, second_info TEXT)"
        )

    def test_set_trade_i(self):
        """Test the set_trade method of the RegistrationDB class where record IS None."""
        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=None
        ) as mock_get_trade_table:
            with patch.object(self.db, "_execute_single_sql") as mock_exe:
                self.db.set_trade(
                    self.address, self.timestamp, self.data,
                )

        # after
        mock_get_trade_table.assert_called_once()
        mock_exe.assert_any_call(
            "INSERT INTO trade_table(address, first_trade, second_trade, first_info, second_info) values(?, ?, ?, ?, ?)",
            (self.address, self.timestamp, None, json.dumps(self.data), None),
        )

    def test_set_trade_ii(self):
        """Test the set_trade method of the RegistrationDB class where record is NOT None."""
        # setup
        record = (
            self.address,
            self.first_trade,
            None,
            self.first_info,
            self.second_info,
        )

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade_table:
            with patch.object(self.db, "_execute_single_sql") as mock_exe:
                self.db.set_trade(
                    self.address, self.timestamp, self.data,
                )

        # after
        mock_get_trade_table.assert_called_once()

        mock_exe.assert_any_call(
            "INSERT or REPLACE into trade_table(address, first_trade, second_trade, first_info, second_info) values(?, ?, ?, ?, ?)",
            (
                self.address,
                self.first_trade,
                self.timestamp,
                self.first_info,
                json.dumps(self.data),
            ),
        )

    def test_set_trade_iii(self):
        """Test the set_trade method of the RegistrationDB class where record is NOT None and first_trade is None."""
        # setup
        record = (
            self.address,
            None,
            None,
            self.first_info,
            self.second_info,
        )

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade_table:
            with patch.object(self.db, "_execute_single_sql") as mock_exe:
                self.db.set_trade(
                    self.address, self.timestamp, self.data,
                )

        # after
        mock_get_trade_table.assert_called_once()

        mock_exe.assert_not_called()

    def test_get_trade_table(self):
        """Test the get_trade_table method of the RegistrationDB class."""
        # setup
        trade_table = ("something_1", "something_2")
        result = [trade_table, ("something_3",)]

        # operation
        with patch.object(
            self.db, "_execute_single_sql", return_value=result
        ) as mock_exe:
            actual_trade_table = self.db.get_trade_table(self.address)

        # after
        mock_exe.assert_any_call(
            "SELECT * FROM trade_table where address=?", (self.address,)
        )
        assert actual_trade_table == trade_table

    def test_set_registered_i(self):
        """Test the set_registered method of the RegistrationDB class where is_registeredis is False."""
        # operation
        with patch.object(
            self.db, "is_registered", return_value=False
        ) as mock_is_registered:
            with patch.object(self.db, "_execute_single_sql") as mock_exe:
                self.db.set_registered(
                    self.address, self.developer_handle,
                )

        # after
        mock_is_registered.assert_called_once()
        mock_exe.assert_any_call(
            "INSERT OR REPLACE INTO registered_table(address, ethereum_address, ethereum_signature, fetchai_signature, developer_handle, tweet) values(?, ?, ?, ?, ?, ?)",
            (self.address, "", "", "", self.developer_handle, ""),
        )

    def test_set_registered_ii(self):
        """Test the set_registered method of the RegistrationDB class where is_registeredis is True."""
        # operation
        with patch.object(
            self.db, "is_registered", return_value=True
        ) as mock_is_registered:
            with patch.object(self.db, "_execute_single_sql") as mock_exe:
                self.db.set_registered(
                    self.address, self.developer_handle,
                )

        # after
        mock_is_registered.assert_called_once()
        mock_exe.assert_not_called()

    def test_is_registered_i(self):
        """Test the is_registered method of the RegistrationDB class where result is NOT empty."""
        # setup
        result = [["1"], ["2", "3"]]

        # operation
        with patch.object(
            self.db, "_execute_single_sql", return_value=result
        ) as mock_exe:
            is_registered = self.db.is_registered(self.address)

        # after
        mock_exe.assert_any_call(
            "SELECT * FROM registered_table WHERE address=?", (self.address,)
        )
        assert is_registered

    def test_is_registered_ii(self):
        """Test the is_registered method of the RegistrationDB class where result is empty."""
        # setup
        result = []

        # operation
        with patch.object(
            self.db, "_execute_single_sql", return_value=result
        ) as mock_exe:
            is_registered = self.db.is_registered(self.address)

        # after
        mock_exe.assert_any_call(
            "SELECT * FROM registered_table WHERE address=?", (self.address,)
        )
        assert not is_registered

    def test_is_allowed_to_trade_i(self):
        """Test the is_allowed_to_trade method of the RegistrationDB class where record IS None."""
        # setup
        mininum_hours_between_txs = 1
        record = None

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            with patch.object(self.db, "_execute_single_sql") as mock_exe:
                is_registered = self.db.is_allowed_to_trade(
                    self.address, mininum_hours_between_txs
                )

        # after
        mock_get_trade.assert_called_once()
        mock_exe.assert_not_called()
        assert is_registered

    def test_is_allowed_to_trade_ii(self):
        """Test the is_allowed_to_trade method of the RegistrationDB class where first_trade and second_trade are NOT present."""
        # setup
        minimum_hours_between_txs = 1
        record = (self.address, None, None, self.first_info, self.second_info)

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            is_registered = self.db.is_allowed_to_trade(
                self.address, minimum_hours_between_txs
            )

        # after
        mock_get_trade.assert_called_once()
        assert is_registered

    def test_is_allowed_to_trade_iii(self):
        """Test the is_allowed_to_trade method of the RegistrationDB class where is_allowed_to_trade_ is True."""
        # setup
        minimum_hours_between_txs = 1
        record = (
            self.address,
            self.first_trade,
            None,
            self.first_info,
            self.second_info,
        )

        mocked_now_greater_than_minimum = "2020-12-22 20:30:00.000000"
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime.strptime(
            mocked_now_greater_than_minimum, "%Y-%m-%d %H:%M:%S.%f"
        )

        # operation
        with patch("datetime.datetime", new=datetime_mock):
            with patch.object(
                self.db, "get_trade_table", return_value=record
            ) as mock_get_trade:
                is_registered = self.db.is_allowed_to_trade(
                    self.address, minimum_hours_between_txs
                )

        # after
        mock_get_trade.assert_called_once()
        assert is_registered

    def test_is_allowed_to_trade_iv(self):
        """Test the is_allowed_to_trade method of the RegistrationDB class where is_allowed_to_trade_ is False."""
        # setup
        minimum_hours_between_txs = 1
        record = (
            self.address,
            self.first_trade,
            None,
            self.first_info,
            self.second_info,
        )

        mocked_now_less_than_minimum = "2020-12-22 18:31:00.000000"

        # operation
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime.strptime(
            mocked_now_less_than_minimum, "%Y-%m-%d %H:%M:%S.%f"
        )
        with patch("datetime.datetime", new=datetime_mock):
            with patch.object(
                self.db, "get_trade_table", return_value=record
            ) as mock_get_trade:
                with patch.object(self.logger, "log") as mock_logger:
                    is_registered = self.db.is_allowed_to_trade(
                        self.address, minimum_hours_between_txs
                    )

        # after
        mock_get_trade.assert_called_once()
        mock_logger.assert_any_call(
            logging.INFO,
            f"Invalid attempt for counterparty={self.address}, not enough time since last trade!",
        )
        assert not is_registered

    def test_is_allowed_to_trade_v(self):
        """Test the is_allowed_to_trade method of the RegistrationDB class where second_trade IS present."""
        # setup
        minimum_hours_between_txs = 1
        record = (
            self.address,
            self.first_trade,
            self.second_trade,
            self.first_info,
            self.second_info,
        )

        mocked_now = "2020-12-22 18:31:00.000000"
        # operation
        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime.strptime(
            mocked_now, "%Y-%m-%d %H:%M:%S.%f"
        )
        with patch("datetime.datetime", new=datetime_mock):
            with patch.object(
                self.db, "get_trade_table", return_value=record
            ) as mock_get_trade:
                with patch.object(self.logger, "log") as mock_logger:
                    is_registered = self.db.is_allowed_to_trade(
                        self.address, minimum_hours_between_txs
                    )

        # after
        mock_get_trade.assert_called_once()
        mock_logger.assert_any_call(
            logging.INFO,
            f"Invalid attempt for counterparty={self.address}, already completed 2 trades!",
        )
        assert not is_registered

    def test_has_completed_two_trades_i(self):
        """Test the has_completed_two_trades method of the RegistrationDB class where first and second trade are present."""
        # setup
        record = (
            self.address,
            self.first_trade,
            self.second_trade,
            self.first_info,
            self.second_info,
        )

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            has_completed = self.db.has_completed_two_trades(self.address)

        # after
        mock_get_trade.assert_called_once()
        assert has_completed

    def test_has_completed_two_trades_ii(self):
        """Test the has_completed_two_trades method of the RegistrationDB class where first and second trade are NOT present."""
        # setup
        record = (self.address, None, None, self.first_info, self.second_info)

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            has_completed = self.db.has_completed_two_trades(self.address)

        # after
        mock_get_trade.assert_called_once()
        assert not has_completed

    def test_has_completed_two_trades_iii(self):
        """Test the has_completed_two_trades method of the RegistrationDB class where first trade is NOT and second trade IS present."""
        # setup
        record = (
            self.address,
            None,
            self.second_trade,
            self.first_info,
            self.second_info,
        )

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            has_completed = self.db.has_completed_two_trades(self.address)

        # after
        mock_get_trade.assert_called_once()
        assert not has_completed

    def test_has_completed_two_trades_iv(self):
        """Test the has_completed_two_trades method of the RegistrationDB class where first trade IS and second trade is NOT present."""
        # setup
        record = (
            self.address,
            self.first_trade,
            None,
            self.first_info,
            self.second_info,
        )

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            has_completed = self.db.has_completed_two_trades(self.address)

        # after
        mock_get_trade.assert_called_once()
        assert not has_completed

    def test_has_completed_two_trades_v(self):
        """Test the has_completed_two_trades method of the RegistrationDB class where record IS None."""
        # setup
        record = None

        # operation
        with patch.object(
            self.db, "get_trade_table", return_value=record
        ) as mock_get_trade:
            has_completed = self.db.has_completed_two_trades(self.address)

        # after
        mock_get_trade.assert_called_once()
        assert not has_completed

    def test_completed_two_trades(self):
        """Test the completed_two_trades method of the RegistrationDB class."""
        # setup
        row_1 = (
            "address_1",
            "ethereum_address_1",
            "something2_1",
            "something3_1",
            "developer_handle_1",
        )
        row_2 = (
            "address_2",
            "ethereum_address_2",
            "something2_2",
            "something3_2",
            "developer_handle_2",
        )
        row_3 = (
            "address_3",
            "ethereum_address_3",
            "something2_3",
            "something3_3",
            "developer_handle_3",
        )
        result = [row_1, row_2, row_3]

        has_completed = [True, False, True]

        # operation
        with patch.object(
            self.db, "_execute_single_sql", return_value=result
        ) as mock_exe:
            with patch.object(
                self.db, "has_completed_two_trades", side_effect=has_completed
            ):
                actual_completed = self.db.completed_two_trades()

        # after
        mock_exe.assert_any_call("SELECT * FROM registered_table", ())
        assert actual_completed == [
            ("address_1", "ethereum_address_1", "developer_handle_1"),
            ("address_3", "ethereum_address_3", "developer_handle_3"),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains class representations corresponding to every custom type in the protocol specification."""

from enum import Enum
from typing import Any, Dict

from aea.exceptions import enforce
from aea.helpers.search.models import Description as BaseDescription
from aea.helpers.search.models import Query as BaseQuery
from aea.helpers.serializers import DictProtobufStructSerializer


Description = BaseDescription


class AgentsInfo:
    """This class represents an instance of AgentsInfo."""

    __slots__ = ("_body",)

    def __init__(self, body: Dict[str, Dict[str, Any]]):
        """Initialise an instance of AgentsInfo."""
        self._body = body
        self._check_consistency()

    def _check_consistency(self) -> None:
        """Check consistency of the object."""
        if self._body is None:
            raise ValueError("body must not be None")
        enforce(
            isinstance(self._body, dict)
            and all([isinstance(key, str) for key in self._body.keys()])
            and all([isinstance(value, dict) for value in self._body.values()]),
            "Body must be dict and keys must be str and values must be dict.",
        )

    @property
    def body(self) -> Dict[str, Dict[str, Any]]:
        """Get the body."""
        return self._body

    def get_info_for_agent(self, agent_address: str) -> Dict[str, Any]:
        """Get the info for the agent address."""
        return self._body.get(agent_address, {})

    @staticmethod
    def encode(
        agents_info_protobuf_object: Any, agents_info_object: "AgentsInfo"
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the agents_info_protobuf_object argument is matched with the instance of this class in the 'agents_info_object' argument.

        :param agents_info_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param agents_info_object: an instance of this class to be encoded in the protocol buffer object.
        :return: None
        """
        agents_info_protobuf_object.agents_info = DictProtobufStructSerializer.encode(
            agents_info_object.body
        )

    @classmethod
    def decode(cls, agents_info_protobuf_object: Any) -> "AgentsInfo":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'agents_info_protobuf_object' argument.

        :param agents_info_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'agents_info_protobuf_object' argument.
        """
        body = DictProtobufStructSerializer.decode(
            agents_info_protobuf_object.agents_info
        )
        return cls(body)

    def __eq__(self, other: Any) -> bool:
        """Compare with another object."""
        if not isinstance(other, AgentsInfo):
            return False  # pragma: nocover
        return self.body == other.body


class OefErrorOperation(Enum):
    """This class represents an instance of OefErrorOperation."""

    REGISTER_SERVICE = 0
    UNREGISTER_SERVICE = 1
    SEARCH_SERVICES = 2
    SEND_MESSAGE = 3

    OTHER = 10000

    def __str__(self) -> str:
        """Get string representation."""
        return str(self.value)

    @staticmethod
    def encode(
        oef_error_operation_protobuf_object: Any,
        oef_error_operation_object: "OefErrorOperation",
    ) -> None:
        """
        Encode an instance of this class into the protocol buffer object.

        The protocol buffer object in the oef_error_operation_protobuf_object argument is matched with the instance of this class in the 'oef_error_operation_object'
        argument.

        :param oef_error_operation_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :param oef_error_operation_object: an instance of this class to be encoded in the protocol buffer object.
        :return: None
        """
        oef_error_operation_protobuf_object.oef_error = oef_error_operation_object.value

    @classmethod
    def decode(cls, oef_error_operation_protobuf_object: Any) -> "OefErrorOperation":
        """
        Decode a protocol buffer object that corresponds with this class into an instance of this class.

        A new instance of this class is created that matches the protocol buffer object in the 'oef_error_operation_protobuf_object' argument.

        :param oef_error_operation_protobuf_object: the protocol buffer object whose type corresponds with this class.
        :return: A new instance of this class that matches the protocol buffer object in the 'oef_error_operation_protobuf_object' argument.
        """
        enum_value_from_pb2 = oef_error_operation_protobuf_object.oef_error
        return OefErrorOperation(enum_value_from_pb2)


Query = BaseQuery



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the integration test for the coin price skill."""

import time
from pathlib import Path
from typing import Dict

import pytest

from aea.helpers import http_requests as requests
from aea.test_tools.test_cases import AEATestCaseEmpty

from tests.conftest import ROOT_DIR


API_SPEC_PATH = str(
    Path(
        ROOT_DIR, "packages", "fetchai", "skills", "coin_price", "coin_api_spec.yaml"
    ).absolute()
)


def parse_prometheus_output(prom_data: bytes) -> Dict[str, float]:
    """Convert prometheus text output to a dict of {"metric": value}"""
    metrics = {}
    for line in prom_data.decode().splitlines():
        tokens = line.split()
        if tokens[0] != "#":
            metrics.update({tokens[0]: float(tokens[1])})
    return metrics


@pytest.mark.integration
class TestCoinPriceSkill(AEATestCaseEmpty):
    """Test that coin price skill works."""

    def test_coin_price(self):
        """Run the coin price skill sequence."""

        coin_price_feed_aea_name = self.agent_name

        self.add_item("connection", "fetchai/http_client:0.18.0")
        self.add_item("connection", "fetchai/http_server:0.17.0")
        self.add_item("connection", "fetchai/prometheus:0.4.0")
        self.add_item("skill", "fetchai/coin_price:0.6.0")
        self.set_config("agent.default_connection", "fetchai/http_server:0.17.0")

        default_routing = {
            "fetchai/http:0.13.0": "fetchai/http_client:0.18.0",
            "fetchai/prometheus:0.4.0": "fetchai/prometheus:0.4.0",
        }
        setting_path = "agent.default_routing"
        self.nested_set_config(setting_path, default_routing)

        self.set_config(
            "vendor.fetchai.connections.http_server.config.api_spec_path", API_SPEC_PATH
        )
        self.set_config(
            "vendor.fetchai.connections.http_server.config.target_skill_id",
            "fetchai/coin_price:0.6.0",
        )
        self.set_config(
            "vendor.fetchai.skills.coin_price.models.coin_price_model.args.use_http_server",
            True,
            type_="bool",
        )

        diff = self.difference_to_fetched_agent(
            "fetchai/coin_price_feed:0.7.0", coin_price_feed_aea_name
        )
        assert (
            diff == []
        ), "Difference between created and fetched project for files={}".format(diff)

        self.run_install()

        process = self.run_agent()
        is_running = self.is_running(process)
        assert is_running, "AEA not running within timeout!"

        time.sleep(6)  # we wait a bit longer than the tick rate of the behaviour

        response = requests.get("http://127.0.0.1:8000/price")
        assert response.status_code == 200, "Failed to get response code 200"
        coin_price = response.content.decode("utf-8")
        assert "value" in coin_price, "Response does not contain 'value'"
        assert "decimals" in coin_price, "Response does not contain 'decimals'"

        response = requests.get("http://127.0.0.1:8000")
        assert response.status_code == 404
        assert response.content == b"", "Get request should not work without valid path"

        response = requests.post("http://127.0.0.1:8000/price")
        assert response.status_code == 404
        assert response.content == b"", "Post not allowed"

        # test prometheus metrics
        prom_response = requests.get("http://127.0.0.1:9090/metrics")
        metrics = parse_prometheus_output(prom_response.content)
        assert metrics["num_retrievals"] > 0.0, "num_retrievals metric not updated"
        assert metrics["num_requests"] == 1.0, "num_requests metric not equal to 1"

        self.terminate_agents()
        assert (
            self.is_successfully_terminated()
        ), "Http echo agent wasn't successfully terminated."



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a simple Fetch oracle contract deployment behaviour."""

from typing import Any, Dict, cast

from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.connections.ledger.base import CONNECTION_ID as LEDGER_API_ADDRESS
from packages.fetchai.connections.prometheus.connection import (
    PUBLIC_ID as PROM_CONNECTION_ID,
)
from packages.fetchai.contracts.oracle.contract import PUBLIC_ID as CONTRACT_PUBLIC_ID
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.prometheus.message import PrometheusMessage
from packages.fetchai.skills.simple_oracle.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    LedgerApiDialogues,
    PrometheusDialogues,
)
from packages.fetchai.skills.simple_oracle.strategy import Strategy


DEFAULT_UPDATE_INTERVAL = 5
EXPIRATION_BLOCK = 1000000000000000


class SimpleOracleBehaviour(TickerBehaviour):
    """This class implements a behaviour that deploys a Fetch oracle contract."""

    def __init__(self, **kwargs: Any):
        """Initialise the behaviour."""
        update_interval = kwargs.pop(
            "update_interval", DEFAULT_UPDATE_INTERVAL
        )  # type: int
        super().__init__(tick_interval=update_interval, **kwargs)

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """

        self.context.logger.info("Setting up Fetch oracle contract...")
        strategy = cast(Strategy, self.context.strategy)

        if not strategy.is_contract_deployed:
            self._request_contract_deploy_transaction()
        else:
            self.context.logger.info("Fetch oracle contract address already added")

        if strategy.is_oracle_role_granted:
            self.context.logger.info("Oracle role already granted")

        prom_dialogues = cast(PrometheusDialogues, self.context.prometheus_dialogues)

        if prom_dialogues.enabled:
            for metric in prom_dialogues.metrics:
                metric_name = metric["name"]
                self.context.logger.info("Adding Prometheus metric: " + metric_name)
                self.add_prometheus_metric(
                    metric_name,
                    metric["type"],
                    metric["description"],
                    dict(metric["labels"]),
                )

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """

        strategy = cast(Strategy, self.context.strategy)

        # Request account balance
        self._get_balance()

        if not strategy.is_contract_deployed:
            self.context.logger.info("Oracle contract not yet deployed")
            return

        if not strategy.is_oracle_role_granted:
            self.context.logger.info("Oracle not yet created")
            self._request_grant_role_transaction()
            return

        # Check for entropy value from fetch oracle skill
        oracle_data = self.context.shared_state.get("oracle_data", None)
        if oracle_data is None:
            self.context.logger.info("No oracle value to publish")
        else:
            self.context.logger.info("Publishing oracle value")

            # add expiration block
            update_args = oracle_data.copy()
            update_args["expiration_block"] = EXPIRATION_BLOCK
            self._request_update_transaction(update_args)

    def _request_contract_deploy_transaction(self) -> None:
        """
        Request contract deployment transaction

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=str(CONTRACT_PUBLIC_ID),
            callable="get_deploy_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.context.agent_address,
                    "ERC20Address": strategy.erc20_address,
                    "initialFee": strategy.initial_fee_deploy,
                    "gas": strategy.default_gas_deploy,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue,)
        contract_api_dialogue.terms = strategy.get_deploy_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting contract deployment transaction...")

    def _request_grant_role_transaction(self) -> None:
        """
        Request transaction that grants oracle role in a Fetch oracle contract

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=str(CONTRACT_PUBLIC_ID),
            contract_address=strategy.contract_address,
            callable="get_grant_role_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "oracle_address": self.context.agent_address,
                    "gas": strategy.default_gas_grant_role,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue)
        contract_api_dialogue.terms = strategy.get_grant_role_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting grant role transaction...")

    def _request_update_transaction(self, update_args: Dict[str, Any]) -> None:
        """
        Request transaction that updates value in Fetch oracle contract

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=str(CONTRACT_PUBLIC_ID),
            contract_address=strategy.contract_address,
            callable="get_update_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "oracle_address": self.context.agent_address,
                    "update_function": strategy.update_function,
                    "update_args": list(update_args.values()),
                    "gas": strategy.default_gas_update,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue)
        contract_api_dialogue.terms = strategy.get_update_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting update transaction...")

    def _get_balance(self) -> None:
        """
        Request balance of agent account by sending a message to the ledger API

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        ledger_api_dialogues = cast(
            LedgerApiDialogues, self.context.ledger_api_dialogues
        )
        ledger_api_msg, _ = ledger_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=strategy.ledger_id,
            address=cast(str, self.context.agent_addresses.get(strategy.ledger_id)),
        )
        self.context.outbox.put_message(message=ledger_api_msg)

    def add_prometheus_metric(
        self,
        metric_name: str,
        metric_type: str,
        description: str,
        labels: Dict[str, str],
    ) -> None:
        """
        Add a prometheus metric.

        :param metric_name: the name of the metric to add.
        :param type: the type of the metric.
        :param description: a description of the metric.
        :param labels: the metric labels.
        :return: None
        """

        # context
        prom_dialogues = cast(PrometheusDialogues, self.context.prometheus_dialogues)

        # prometheus update message
        message, _ = prom_dialogues.create(
            counterparty=str(PROM_CONNECTION_ID),
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type=metric_type,
            title=metric_name,
            description=description,
            labels=labels,
        )

        # send message
        self.context.outbox.put_message(message=message)

    def update_prometheus_metric(
        self, metric_name: str, update_func: str, value: float, labels: Dict[str, str],
    ) -> None:
        """
        Update a prometheus metric.

        :param metric_name: the name of the metric.
        :param update_func: the name of the update function (e.g. inc, dec, set, ...).
        :param value: the value to provide to the update function.
        :param labels: the metric labels.
        :return: None
        """

        # context
        prom_dialogues = cast(PrometheusDialogues, self.context.prometheus_dialogues)

        # prometheus update message
        message, _ = prom_dialogues.create(
            counterparty=str(PROM_CONNECTION_ID),
            performative=PrometheusMessage.Performative.UPDATE_METRIC,
            title=metric_name,
            callable=update_func,
            value=value,
            labels=labels,
        )

        # send message
        self.context.outbox.put_message(message=message)

    def teardown(self) -> None:
        """
        Implement the task teardown.

        :return: None
        """



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains testing utilities."""
import logging
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

import docker
import requests
from docker import DockerClient
from docker.models.containers import Container

from aea.exceptions import enforce


logger = logging.getLogger(__name__)


class DockerImage(ABC):
    """A class to wrap interatction with a Docker image."""

    def __init__(self, client: docker.DockerClient):
        """Initialize."""
        self._client = client

    def check_skip(self):
        """
        Check whether the test should be skipped.

        By default, nothing happens.
        """

    @property
    @abstractmethod
    def tag(self) -> str:
        """Return the tag of the image."""

    def stop_if_already_running(self):
        """Stop the running images with the same tag, if any."""
        client = docker.from_env()
        for container in client.containers.list():
            if self.tag in container.image.tags:
                logger.info(f"Stopping image {self.tag}...")
                container.stop()

    @abstractmethod
    def create(self) -> Container:
        """Instantiate the image in a container."""

    @abstractmethod
    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """
        Wait until the image is running.

        :param max_attempts: max number of attempts.
        :param sleep_rate: the amount of time to sleep between different requests.
        :return: True if the wait was successful, False otherwise.
        """
        return True


class GanacheDockerImage(DockerImage):
    """Wrapper to Ganache Docker image."""

    def __init__(
        self,
        client: DockerClient,
        addr: str,
        port: int,
        config: Optional[Dict] = None,
        gas_limit: int = 10000000000000,
    ):
        """
        Initialize the Ganache Docker image.

        :param client: the Docker client.
        :param addr: the address.
        :param port: the port.
        :param config: optional configuration to command line.
        """
        super().__init__(client)
        self._addr = addr
        self._port = port
        self._config = config or {}
        self._gas_limit = gas_limit

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return "trufflesuite/ganache-cli:latest"

    def _make_ports(self) -> Dict:
        """Make ports dictionary for Docker."""
        return {f"{self._port}/tcp": ("0.0.0.0", self._port)}  # nosec

    def _build_command(self) -> List[str]:
        """Build command."""
        cmd = ["ganache-cli"]
        cmd += ["--gasLimit=" + str(self._gas_limit)]
        accounts_balances = self._config.get("accounts_balances", [])
        for account, balance in accounts_balances:
            cmd += [f"--account='{account},{balance}'"]
        return cmd

    def create(self) -> Container:
        """Create the container."""
        cmd = self._build_command()
        container = self._client.containers.run(
            self.tag, command=cmd, detach=True, ports=self._make_ports()
        )
        return container

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""
        request = dict(jsonrpc=2.0, method="web3_clientVersion", params=[], id=1)
        for i in range(max_attempts):
            try:
                response = requests.post(f"{self._addr}:{self._port}", json=request)
                enforce(response.status_code == 200, "")
                return True
            except Exception:
                logger.info(
                    "Attempt %s failed. Retrying in %s seconds...", i, sleep_rate
                )
                time.sleep(sleep_rate)
        return False



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains sqlite storage backend implementation."""
import asyncio
import json
import os
import platform
import sqlite3
import sys
import threading
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import List, Optional, Tuple
from urllib.parse import urlparse

from aea.helpers.storage.backends.base import (
    AbstractStorageBackend,
    EQUALS_TYPE,
    JSON_TYPES,
    OBJECT_ID_AND_BODY,
)


class SqliteStorageBackend(AbstractStorageBackend):
    """Sqlite storage backend."""

    def __init__(self, uri: str) -> None:
        """Init backend."""
        super().__init__(uri)
        parsed = urlparse(self._uri)
        self._fname = parsed.netloc or parsed.path
        self._connection: Optional[sqlite3.Connection] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._lock = threading.Lock()
        self._executor = ThreadPoolExecutor(max_workers=1)

    def _execute_sql_sync(self, query: str, args: Optional[List] = None) -> List[Tuple]:
        """
        Execute sql command and return results.

        :param query: sql query string
        :param args: optional argumets to set into sql query.

        :return: List of tuples with sql records
        """
        if not self._connection:  # pragma: nocover
            raise ValueError("Not connected")
        with self._lock:
            result = self._connection.execute(query, args or []).fetchall()
            self._connection.commit()
            return result

    async def _executute_sql(
        self, query: str, args: Optional[List] = None
    ) -> Optional[JSON_TYPES]:
        """
        Execute sql command and return results in async executor.

        :param query: sql query string
        :param args: optional argumets to set into sql query.

        :return: List of tuples with sql records
        """
        if not self._loop:  # pragma: nocover
            raise ValueError("Not connected")
        return await self._loop.run_in_executor(
            self._executor, self._execute_sql_sync, query, args
        )

    async def connect(self) -> None:
        """Connect to backend."""
        self._loop = asyncio.get_event_loop()
        self._connection = await self._loop.run_in_executor(
            self._executor, self._do_connect, self._fname
        )

    @staticmethod
    def _do_connect(fname: str) -> sqlite3.Connection:
        con = sqlite3.connect(fname)
        if (
            platform.system() == "Windows"
            and sys.version_info.major == 3
            and sys.version_info.minor < 9
        ):  # pragma: nocover
            con.enable_load_extension(True)
            path_ext = Path(
                os.path.join(os.path.dirname(__file__), "binaries", "json1.dll")
            ).as_posix()
            con.load_extension(path_ext)
        return con

    async def disconnect(self) -> None:
        """Disconnect the backend."""
        if not self._loop or not self._connection:  # pragma: nocover
            raise ValueError("Not connected")
        await self._loop.run_in_executor(self._executor, self._connection.close)
        self._connection = None
        self._loop = None

    async def ensure_collection(self, collection_name: str) -> None:
        """
        Create collection if not exits.

        :param collection_name: str.
        :return: None
        """
        self._check_collection_name(collection_name)
        sql = f"""CREATE TABLE IF NOT EXISTS {collection_name} (
            object_id TEXT PRIMARY KEY,
            object_body JSON1 NOT NULL)
        """  # nosec
        await self._executute_sql(sql)

    async def put(
        self, collection_name: str, object_id: str, object_body: JSON_TYPES
    ) -> None:
        """
        Put object into collection.

        :param collection_name: str.
        :param object_id: str object id
        :param object_body: python dict, json compatible.
        :return: None
        """
        self._check_collection_name(collection_name)
        sql = f"""INSERT OR REPLACE INTO {collection_name} (object_id, object_body)
            VALUES (?, ?);
        """  # nosec
        await self._executute_sql(sql, [object_id, json.dumps(object_body)])

    async def get(self, collection_name: str, object_id: str) -> Optional[JSON_TYPES]:
        """
        Get object from the collection.

        :param collection_name: str.
        :param object_id: str object id

        :return: dict if object exists in collection otherwise None
        """
        self._check_collection_name(collection_name)
        sql = f"""SELECT object_body FROM {collection_name} WHERE object_id = ? LIMIT 1;"""  # nosec
        result = await self._executute_sql(sql, [object_id])
        if (
            result
            and isinstance(result, (list, tuple))
            and len(result) > 0
            and isinstance(result[0], (list, tuple))
            and len(result[0]) > 0
        ):
            return json.loads(result[0][0])
        return None

    async def remove(self, collection_name: str, object_id: str) -> None:
        """
        Remove object from the collection.

        :param collection_name: str.
        :param object_id: str object id

        :return: None
        """
        self._check_collection_name(collection_name)
        sql = f"""DELETE FROM {collection_name} WHERE object_id = ?;"""  # nosec
        await self._executute_sql(sql, [object_id])

    async def find(
        self, collection_name: str, field: str, equals: EQUALS_TYPE
    ) -> List[OBJECT_ID_AND_BODY]:
        """
        Get objects from the collection by filtering by field value.

        :param collection_name: str.
        :param field: field name to search: example "parent.field"
        :param equals: value field should be equal to

        :return: None
        """
        self._check_collection_name(collection_name)
        sql = f"""SELECT object_id, object_body FROM {collection_name} WHERE json_extract(object_body, ?) = ?;"""  # nosec
        if not field.startswith("$."):
            field = f"$.{field}"
        return [
            (i[0], json.loads(i[1]))
            for i in await self._executute_sql(sql, [field, equals])  # type: ignore
        ]

    async def list(self, collection_name: str) -> List[OBJECT_ID_AND_BODY]:
        """
        List all objects with keys from the collection.

        :param collection_name: str.
        :return: Tuple of objects keys, bodies.
        """
        self._check_collection_name(collection_name)
        sql = f"""SELECT object_id, object_body FROM {collection_name};"""  # nosec
        return [(i[0], json.loads(i[1])) for i in await self._executute_sql(sql)]  # type: ignore



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains handlers for the fetch_beacon skill."""

import json
from typing import Any, Optional, cast

from vyper.utils import keccak256

from aea.protocols.base import Message
from aea.skills.base import Handler

from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.fetch_beacon.dialogues import HttpDialogue, HttpDialogues


class HttpHandler(Handler):
    """This class provides a simple http handler."""

    SUPPORTED_PROTOCOL = HttpMessage.protocol_id

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the handler."""
        super().__init__(**kwargs)
        self.received_http_count = 0

    def setup(self) -> None:
        """Set up the handler."""
        self.context.logger.info("setting up HttpHandler")

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to a message.

        :param message: the message
        :return: None
        """

        message = cast(HttpMessage, message)

        # recover dialogue
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)
        http_dialogue = cast(HttpDialogue, http_dialogues.update(message))
        if http_dialogue is None:
            self._handle_unidentified_dialogue(message)
            return

        if (
            message.performative == HttpMessage.Performative.RESPONSE
            and message.status_code == 200
        ):
            self._handle_response(message)
        else:  # pragma: nocover
            self.context.logger.info(
                "got unexpected http response: code = " + str(message.status_code)
            )

    def _handle_response(self, message: HttpMessage) -> None:
        """
        Handle an http response.

        :param msg: the http message to be handled
        :return: None
        """

        msg_body = json.loads(message.body)

        # get entropy and block data
        entropy = (
            msg_body.get("result", {})
            .get("block", {})
            .get("header", {})
            .get("entropy", {})
            .get("group_signature", None)
        )
        block_hash = msg_body.get("result", {}).get("block_id", {}).get("hash", {})
        block_height_str = (
            msg_body.get("result", {})
            .get("block", {})
            .get("header", {})
            .get("height", None)
        )

        if block_height_str:
            block_height = int(block_height_str)  # type: Optional[int]
        else:
            block_height = None

        if entropy is None:
            self.context.logger.info("entropy not present")
        elif block_height is None:  # pragma: nocover
            self.context.logger.info("block height not present")
        else:
            beacon_data = {
                "entropy": keccak256(entropy.encode("utf-8")),
                "block_hash": bytes.fromhex(block_hash),
                "block_height": block_height,
            }
            self.context.logger.info(
                "Beacon info: "
                + str({"block_height": block_height, "entropy": entropy})
            )
            self.context.shared_state["oracle_data"] = beacon_data

    def _handle_unidentified_dialogue(self, msg: Message) -> None:
        """
        Handle an unidentified dialogue.

        :param msg: the unidentified message to be handled
        :return: None
        """

        self.context.logger.info(
            "received invalid message={}, unidentified dialogue.".format(msg)
        )

    def teardown(self) -> None:
        """
        Teardown the handler.

        :return: None
        """
        self.context.logger.info("tearing down HttpHandler")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the implementation of AEA agents project configuiration."""
import os
from copy import deepcopy
from pathlib import Path
from shutil import rmtree
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from aea.aea import AEA
from aea.aea_builder import AEABuilder
from aea.cli.fetch import do_fetch
from aea.cli.issue_certificates import issue_certificates_
from aea.cli.utils.context import Context
from aea.configurations.base import AgentConfig, PublicId
from aea.configurations.constants import DEFAULT_REGISTRY_NAME
from aea.configurations.data_types import ComponentId
from aea.configurations.manager import AgentConfigManager
from aea.crypto.helpers import create_private_key, get_wallet_from_agent_config


class _Base:
    """Base class to share some methods."""

    @classmethod
    def _get_agent_config(cls, path: Union[Path, str]) -> AgentConfig:
        """Get agent config instance."""
        agent_config = AEABuilder.try_to_load_agent_configuration_file(path)
        agent_config.check_aea_version()
        return agent_config

    @classmethod
    def _get_builder(
        cls,
        agent_config: AgentConfig,
        aea_project_path: Union[Path, str],
        skip_consistency_check: bool = False,
    ) -> AEABuilder:
        """Get AEABuilder instance."""
        builder = AEABuilder(
            with_default_packages=False, build_dir_root=str(aea_project_path)
        )
        builder.set_from_configuration(
            agent_config, Path(aea_project_path), skip_consistency_check
        )
        return builder

    @property
    def builder(self) -> AEABuilder:
        """Get AEABuilder instance."""
        raise NotImplementedError  # pragma: nocover

    def install_pypi_dependencies(self) -> None:
        """Install python dependencies for the project."""
        self.builder.install_pypi_dependencies()


class Project(_Base):
    """Agent project representation."""

    __slots__ = ("public_id", "path", "agents")

    def __init__(self, public_id: PublicId, path: str) -> None:
        """Init project with public_id and project's path."""
        self.public_id: PublicId = public_id
        self.path: str = path
        self.agents: Set[str] = set()

    def build(self) -> None:
        """Call all build entry points."""
        self.builder.call_all_build_entrypoints()

    @classmethod
    def load(
        cls,
        working_dir: str,
        public_id: PublicId,
        is_local: bool = False,
        is_remote: bool = False,
        is_restore: bool = False,
        registry_path: str = DEFAULT_REGISTRY_NAME,
        skip_consistency_check: bool = False,
    ) -> "Project":
        """
        Load project with given public_id to working_dir.

        If local = False and remote = False, then the packages
        are fetched in mixed mode (i.e. first try from local
        registry, and then from remote registry in case of failure).

        :param working_dir: the working directory
        :param public_id: the public id
        :param is_local: whether to fetch from local
        :param is_remote whether to fetch from remote
        :param registry_path: the path to the registry locally
        :param skip_consistency_check: consistency checks flag
        """
        ctx = Context(cwd=working_dir, registry_path=registry_path)
        ctx.set_config("skip_consistency_check", skip_consistency_check)

        path = os.path.join(working_dir, public_id.author, public_id.name)
        target_dir = os.path.join(public_id.author, public_id.name)

        if not is_restore and not os.path.exists(target_dir):
            do_fetch(ctx, public_id, is_local, is_remote, target_dir=target_dir)
        return cls(public_id, path)

    def remove(self) -> None:
        """Remove project, do cleanup."""
        rmtree(self.path)

    @property
    def builder(self) -> AEABuilder:
        """Get builder instance."""
        return self._get_builder(self._get_agent_config(self.path), self.path)

    def check(self) -> None:
        """Check we can still construct an AEA from the project with builder.build."""
        _ = self.builder


class AgentAlias(_Base):
    """Agent alias representation."""

    __slots__ = ("project", "agent_name", "_data_dir", "_agent_config")

    def __init__(
        self, project: Project, agent_name: str, data_dir: str,
    ):
        """Init agent alias with project, config, name, agent, builder."""
        self.project = project
        self.agent_name = agent_name
        self._data_dir = data_dir
        if not os.path.exists(self._data_dir):
            os.makedirs(self._data_dir)
        self._agent_config: AgentConfig = self._get_agent_config(project.path)
        self._ensure_private_keys()

    def set_agent_config_from_data(self, json_data: List[Dict]) -> None:
        """
        Set agent config instance constructed from json data.

        :param json_data: agent config json data

        :return: None
        """
        self._agent_config = AEABuilder.loader.load_agent_config_from_json(json_data)
        self._ensure_private_keys()

    def _ensure_private_keys(self) -> None:
        """Add proviate keys of not present in the config."""
        builder = self._get_builder(self.agent_config, self.project.path)
        default_ledger = builder.get_default_ledger()

        if not self.agent_config.private_key_paths.read_all():
            self.agent_config.private_key_paths.create(
                default_ledger, self._create_private_key(default_ledger)
            )

        if not self.agent_config.connection_private_key_paths.read_all():
            self.agent_config.connection_private_key_paths.create(
                default_ledger,
                self._create_private_key(default_ledger, is_connection=True),
            )

    @property
    def builder(self) -> AEABuilder:
        """Get builder instance."""
        builder = self._get_builder(self.agent_config, self.project.path)
        builder.set_name(self.agent_name)
        builder.set_runtime_mode("threaded")
        builder.set_data_dir(self._data_dir)
        return builder

    @property
    def agent_config(self) -> AgentConfig:
        """Get agent config."""
        return self._agent_config

    def _create_private_key(
        self, ledger: str, replace: bool = False, is_connection: bool = False
    ) -> str:
        """
        Create new key for agent alias in working dir keys dir.

        If file exists, check `replace` option.
        """
        file_name = (
            f"{ledger}_connection_private.key"
            if is_connection
            else f"{ledger}_private.key"
        )
        filepath = os.path.join(self._data_dir, file_name)
        if os.path.exists(filepath) and not replace:
            return filepath
        create_private_key(ledger, filepath)
        return filepath

    def remove_from_project(self) -> None:
        """Remove agent alias from project."""
        self.project.agents.remove(self.agent_name)

    @property
    def dict(self) -> Dict[str, Any]:
        """Convert AgentAlias to dict."""
        return {
            "public_id": str(self.project.public_id),
            "agent_name": self.agent_name,
            "config": self.config_json,
        }

    @property
    def config_json(self) -> List[Dict]:
        """Get agent config json data."""
        json_data = self.agent_config.ordered_json
        result: List[Dict] = [json_data] + json_data.pop("component_configurations", {})
        return result

    def get_aea_instance(self) -> AEA:
        """Build new aea instance."""
        self.issue_certificates()
        aea = self.builder.build()
        # override build dir to project's one
        aea.DEFAULT_BUILD_DIR_NAME = os.path.join(
            self.project.path, aea.DEFAULT_BUILD_DIR_NAME
        )
        return aea

    def issue_certificates(self) -> None:
        """Issue the certificates for this agent."""
        issue_certificates_(
            self.project.path, self.agent_config_manager, self._data_dir
        )

    def set_overrides(
        self,
        agent_overrides: Optional[Dict] = None,
        component_overrides: Optional[List[Dict]] = None,
    ) -> None:
        """Set override for this agent alias's config."""
        overrides = deepcopy(agent_overrides or {})
        component_configurations: Dict[ComponentId, Dict] = {}

        for component_override in deepcopy(component_overrides or []):
            try:
                component_id = ComponentId.from_json(
                    {"version": "any", **component_override}
                )
                component_override.pop("author")
                component_override.pop("name")
                component_override.pop("type")
                component_override.pop("version")
                component_configurations[component_id] = component_override
            except (ValueError, KeyError) as e:  # pragma: nocover
                raise ValueError(
                    f"Component overrides are incorrect: {e} during process: {component_override}"
                )

        overrides["component_configurations"] = component_configurations
        return self.agent_config_manager.update_config(overrides)

    @property
    def agent_config_manager(self) -> AgentConfigManager:
        """Get agent configuration manager instance for the config."""
        return AgentConfigManager(self.agent_config, self.project.path)

    def get_overridables(self) -> Tuple[Dict, List[Dict]]:
        """Get all overridables for this agent alias's config."""
        (
            agent_overridables,
            components_overridables,
        ) = self.agent_config_manager.get_overridables()
        components_configurations = []
        for component_id, obj in components_overridables.items():
            if not obj:  # pragma: nocover
                continue
            obj.update(component_id.json)
            components_configurations.append(obj)

        return agent_overridables, components_configurations

    def get_addresses(self) -> Dict[str, str]:
        """
        Get addresses from private keys.

        :return: dict with crypto id str as key and address str as value
        """
        wallet = get_wallet_from_agent_config(self.agent_config)
        return wallet.addresses

    def get_connections_addresses(self) -> Dict[str, str]:
        """
        Get connections addresses from connections private keys.

        :return: dict with crypto id str as key and address str as value
        """
        wallet = get_wallet_from_agent_config(self.agent_config)
        return wallet.connection_cryptos.addresses



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the simple_data_request skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.helpers.transaction.base import Terms
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.connections.ledger.base import (
    CONNECTION_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.fetchai.protocols.contract_api.custom_types import Kwargs
from packages.fetchai.protocols.contract_api.custom_types import (
    Kwargs as ContractApiKwargs,
)
from packages.fetchai.protocols.contract_api.custom_types import State
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.ledger_api.custom_types import (
    SignedTransaction,
    TransactionDigest,
    TransactionReceipt,
)
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.prometheus.message import PrometheusMessage
from packages.fetchai.protocols.signing.custom_types import RawTransaction
from packages.fetchai.protocols.signing.message import SigningMessage
from packages.fetchai.skills.simple_oracle.behaviours import SimpleOracleBehaviour
from packages.fetchai.skills.simple_oracle.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    PrometheusDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.fetchai.skills.simple_oracle.handlers import (
    ContractApiHandler,
    LedgerApiHandler,
    PrometheusHandler,
    SigningHandler,
)
from packages.fetchai.skills.simple_oracle.strategy import Strategy

from tests.conftest import ROOT_DIR


LEDGER_ID = "ethereum"
DEFAULT_TX = {"some_tx_key": "some_tx_value"}
DEFAULT_TERMS = [
    LEDGER_ID,
    "sender_address",
    "counterparty_address",
    {},
    {},
    "some_nonce",
]


class TestLedgerApiHandler(BaseSkillTestCase):
    """Test ledger_api handler of simple_oracle skill."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "simple_oracle")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup(**kwargs)
        cls.ledger_api_handler = cast(
            LedgerApiHandler, cls._skill.skill_context.handlers.ledger_api
        )
        cls.logger = cls._skill.skill_context.logger
        cls.simple_oracle_behaviour = cast(
            SimpleOracleBehaviour,
            cls._skill.skill_context.behaviours.simple_oracle_behaviour,
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )

        cls.list_of_ledger_api_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_BALANCE,
                {"ledger_id": LEDGER_ID, "address": "some_eth_address"},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                {"signed_transaction": SignedTransaction(LEDGER_ID, DEFAULT_TX)},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                {"transaction_digest": TransactionDigest(LEDGER_ID, "some_digest")},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.GET_STATE,
                {
                    "ledger_id": LEDGER_ID,
                    "callable": "some_callable",
                    "args": (),
                    "kwargs": LedgerApiMessage.Kwargs({}),
                },
            ),
        )
        cls.list_of_contract_api_messages = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                {
                    "ledger_id": LEDGER_ID,
                    "contract_id": "some_contract_id",
                    "callable": "some_callable",
                    "kwargs": ContractApiKwargs({"some_key": "some_value"}),
                },
            ),
        )
        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {
                    "terms": Terms(*DEFAULT_TERMS),
                    "raw_transaction": RawTransaction(LEDGER_ID, DEFAULT_TX),
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the ledger_api handler."""
        assert self.ledger_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle__handle_unidentified_dialogue(self):
        """Test handling an unidentified dialogoue"""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=LEDGER_ID,
            address="some_eth_address",
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid ledger_api message={incoming_message}, unidentified dialogue.",
        )

        self.assert_quantity_in_outbox(0)

    def test__handle_balance(self):
        """Test handling a balance"""
        # setup
        balance = 0
        dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[:1]
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            performative=LedgerApiMessage.Performative.BALANCE,
            ledger_id=LEDGER_ID,
            balance=balance,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"Balance on {LEDGER_ID} ledger={balance}.",
        )

        self.assert_quantity_in_outbox(1)

    def test__handle_transaction_digest(self):
        """Test handling a transaction digest"""
        # setup
        dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[1:2]
        )
        digest = "some_digest"
        transaction_digest = TransactionDigest(LEDGER_ID, digest)
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
            transaction_digest=transaction_digest,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully submitted. Transaction digest=TransactionDigest: ledger_id={LEDGER_ID}, body={digest}",
        )

        self.assert_quantity_in_outbox(1)

    def test__handle_transaction_receipt_failed(self):
        """Test handling a transaction receipt"""
        # setup
        dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[2:3]
        )

        receipt = {"status": 0}
        transaction_receipt = TransactionReceipt(LEDGER_ID, receipt, DEFAULT_TX)
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
            transaction_receipt=transaction_receipt,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.ERROR,
            f"transaction failed. Transaction receipt={transaction_receipt}",
        )

        self.assert_quantity_in_outbox(0)

    def test__handle_transaction_receipt_deploy(self):
        """Test handling a deploy transaction receipt"""
        # setup
        ledger_api_dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[2:3]
        )
        contract_api_dialogue = self.prepare_skill_dialogue(
            self.contract_api_dialogues, self.list_of_contract_api_messages[:1]
        )
        signing_dialogue = self.prepare_skill_dialogue(
            self.signing_dialogues, self.list_of_signing_messages[:1]
        )

        terms = Terms(*DEFAULT_TERMS, label="deploy")

        contract_api_dialogue.terms = terms
        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        receipt = {"status": 1, "contractAddress": "some_contract_address"}
        transaction_receipt = TransactionReceipt(LEDGER_ID, receipt, DEFAULT_TX)
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=ledger_api_dialogue,
            performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
            transaction_receipt=transaction_receipt,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully settled. Transaction receipt={transaction_receipt}",
        )
        mock_logger.assert_any_call(
            logging.INFO, "Oracle contract successfully deployed!",
        )

        assert (
            self.simple_oracle_behaviour.context.strategy.is_contract_deployed
        ), "Contract deployment status not set"
        self.assert_quantity_in_outbox(0)

    def test__handle_transaction_receipt_grant_role(self):
        """Test handling a grant_role transaction receipt"""
        # setup
        ledger_api_dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[2:3]
        )
        contract_api_dialogue = self.prepare_skill_dialogue(
            self.contract_api_dialogues, self.list_of_contract_api_messages[:1]
        )
        signing_dialogue = self.prepare_skill_dialogue(
            self.signing_dialogues, self.list_of_signing_messages[:1]
        )

        terms = Terms(*DEFAULT_TERMS, label="grant_role")

        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.is_contract_deployed = True

        contract_api_dialogue.terms = terms
        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        receipt = {"status": 1, "contractAddress": "some_contract_address"}
        transaction_receipt = TransactionReceipt(LEDGER_ID, receipt, DEFAULT_TX)
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=ledger_api_dialogue,
            performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
            transaction_receipt=transaction_receipt,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully settled. Transaction receipt={transaction_receipt}",
        )
        mock_logger.assert_any_call(
            logging.INFO, "Oracle role successfully granted!",
        )

        assert (
            self.simple_oracle_behaviour.context.strategy.is_oracle_role_granted
        ), "Oracle role status not set"

        self.assert_quantity_in_outbox(0)

    def test__handle_transaction_receipt_update(self):
        """Test handling an update transaction receipt"""
        # setup
        ledger_api_dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[2:3]
        )
        contract_api_dialogue = self.prepare_skill_dialogue(
            self.contract_api_dialogues, self.list_of_contract_api_messages[:1]
        )
        signing_dialogue = self.prepare_skill_dialogue(
            self.signing_dialogues, self.list_of_signing_messages[:1]
        )

        terms = Terms(*DEFAULT_TERMS, label="update")

        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.is_contract_deployed = True
        strategy.is_oracle_role_granted = True

        contract_api_dialogue.terms = terms
        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        receipt = {"status": 1, "contractAddress": "some_contract_address"}
        transaction_receipt = TransactionReceipt(LEDGER_ID, receipt, DEFAULT_TX)
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=ledger_api_dialogue,
            performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
            transaction_receipt=transaction_receipt,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully settled. Transaction receipt={transaction_receipt}",
        )
        mock_logger.assert_any_call(
            logging.INFO, "Oracle value successfully updated!",
        )

        self.assert_quantity_in_outbox(1)
        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.UPDATE_METRIC,
            callable="inc",
            title="num_oracle_updates",
            value=1.0,
            labels={},
        )
        assert has_attributes, error_str

    def test__handle_error(self):
        """Test handling an error message"""
        # setup
        dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[2:3]
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue, performative=LedgerApiMessage.Performative.ERROR, code=1,
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received ledger_api error message={incoming_message} in dialogue={dialogue}.",
        )

        self.assert_quantity_in_outbox(0)

    def test__handle_invalid(self):
        """Test handling an invalid performative"""
        # setup
        dialogue = self.prepare_skill_dialogue(
            self.ledger_api_dialogues, self.list_of_ledger_api_messages[3:4]
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            performative=LedgerApiMessage.Performative.STATE,
            ledger_id=LEDGER_ID,
            state=LedgerApiMessage.State(LEDGER_ID, {}),
        )

        # operation
        with patch.object(self.ledger_api_handler.context.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle ledger_api message of performative={incoming_message.performative} in dialogue={dialogue}.",
        )

        self.assert_quantity_in_outbox(0)

    def test_teardown(self):
        """Test the teardown method of the ledger_api handler."""
        assert self.ledger_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestContractApiHandler(BaseSkillTestCase):
    """Test contract_api handler of simple oracle."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "simple_oracle")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.contract_api_handler = cast(
            ContractApiHandler, cls._skill.skill_context.handlers.contract_api
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)
        cls.logger = cls.contract_api_handler.context.logger

        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )

        cls.contract_id = "some_contract_id"
        cls.contract_address = "some_contract_address,"
        cls.callable = "some_callable"
        cls.kwargs = Kwargs({"some_key": "some_value"})

        cls.state = State("some_ledger_id", {"some_key": "some_value"})
        cls.terms = Terms(
            ledger_id="some_ledger_id",
            sender_address="some_sender_address",
            counterparty_address="some_counterparty",
            amount_by_currency_id={"1": -10},
            quantities_by_good_id={},
            is_sender_payable_tx_fee=True,
            nonce="some_none",
            fee_by_currency_id={"1": 100},
        )
        cls.list_of_contract_api_messages = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                {
                    "ledger_id": LEDGER_ID,
                    "contract_id": cls.contract_id,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )
        cls.info = {"ethereum_address": "some_ethereum_address"}

    def test_setup(self):
        """Test the setup method of the contract_api handler."""
        assert self.contract_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the contract_api handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=ContractApiMessage.Performative.STATE,
            state=self.state,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid contract_api message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_raw_transaction(self):
        """Test the _handle_raw_transaction method of the contract_api handler."""
        # setup
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.contract_api_dialogues,
                messages=self.list_of_contract_api_messages[:1],
            ),
        )
        contract_api_dialogue.terms = self.terms
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=contract_api_dialogue,
            performative=ContractApiMessage.Performative.RAW_TRANSACTION,
            raw_transaction=ContractApiMessage.RawTransaction(LEDGER_ID, {}),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"received raw transaction={incoming_message}",
        )
        mock_logger.assert_any_call(
            logging.INFO,
            "proposing the transaction to the decision maker. Waiting for confirmation ...",
        )

        self.assert_quantity_in_decision_making_queue(1)

    def test_handle_error(self):
        """Test the _handle_error method of the contract_api handler."""
        # setup
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.contract_api_dialogues,
                messages=self.list_of_contract_api_messages[:1],
            ),
        )

        incoming_message = cast(
            ContractApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=contract_api_dialogue,
                performative=ContractApiMessage.Performative.ERROR,
                data=b"some_data",
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received ledger_api error message={incoming_message} in dialogue={contract_api_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the contract_api handler."""
        # setup
        invalid_performative = ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            ledger_id=LEDGER_ID,
            contract_id=self.contract_id,
            callable=self.callable,
            kwargs=self.kwargs,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle contract_api message of performative={invalid_performative} in dialogue={self.contract_api_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the contract_api handler."""
        assert self.contract_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestSigningHandler(BaseSkillTestCase):
    """Test signing handler of simple oracle."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "simple_oracle")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.signing_handler = cast(
            SigningHandler, cls._skill.skill_context.handlers.signing
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.logger = cls.signing_handler.context.logger

        cls.terms = Terms(
            "some_ledger_id",
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {
                    "terms": cls.terms,
                    "raw_transaction": SigningMessage.RawTransaction(
                        "some_ledger_id", {"some_key": "some_value"}
                    ),
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the signing handler."""
        assert self.signing_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the signing handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=SigningMessage.Performative.ERROR,
            error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid signing message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_signed_transaction(self,):
        """Test the _handle_signed_transaction method of the signing handler."""
        # setup
        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:1],
                counterparty=signing_counterparty,
            ),
        )
        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    "some_ledger_id", {"some_key": "some_value"}
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, "transaction signing was successful.")

        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            to=str(LEDGER_CONNECTION_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            signed_transaction=incoming_message.signed_transaction,
        )
        assert has_attributes, error_str

        assert (
            cast(
                LedgerApiDialogue, self.ledger_api_dialogues.get_dialogue(message)
            ).associated_signing_dialogue
            == signing_dialogue
        )

        mock_logger.assert_any_call(logging.INFO, "sending transaction to ledger.")

    def test_handle_error(self):
        """Test the _handle_error method of the signing handler."""
        # setup
        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = self.prepare_skill_dialogue(
            dialogues=self.signing_dialogues,
            messages=self.list_of_signing_messages[:1],
            counterparty=signing_counterparty,
        )
        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.ERROR,
                error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction signing was not successful. Error_code={incoming_message.error_code} in dialogue={signing_dialogue}",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the signing handler."""
        # setup
        invalid_performative = SigningMessage.Performative.SIGN_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            terms=self.terms,
            raw_transaction=SigningMessage.RawTransaction(
                "some_ledger_id", {"some_key": "some_value"}
            ),
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle signing message of performative={invalid_performative} in dialogue={self.signing_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the signing handler."""
        assert self.signing_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestPrometheusHandler(BaseSkillTestCase):
    """Test prometheus handler of simple_oracle skill."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "simple_oracle")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup()
        cls.prometheus_handler = cast(
            PrometheusHandler, cls._skill.skill_context.handlers.prometheus
        )
        cls.logger = cls._skill.skill_context.logger

        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)

        cls.prometheus_dialogues = cast(
            PrometheusDialogues, cls._skill.skill_context.prometheus_dialogues
        )

        cls.list_of_messages = (
            DialogueMessage(
                PrometheusMessage.Performative.ADD_METRIC,
                {
                    "type": "Gauge",
                    "title": "some_title",
                    "description": "some_description",
                    "labels": {},
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the prometheus handler."""
        assert self.prometheus_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_response(self):
        """Test the _handle_response method of the prometheus handler to a valid response."""
        # setup
        prometheus_dialogue = self.prepare_skill_dialogue(
            dialogues=self.prometheus_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=prometheus_dialogue,
            performative=PrometheusMessage.Performative.RESPONSE,
            code=200,
            message="some_message",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.prometheus_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG, "Prometheus response (200): some_message"
        )

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_response method of the prometheus handler to an unidentified dialogue."""
        # setup
        incoming_message = self.build_incoming_message(
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.RESPONSE,
            code=200,
            message="some_message",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.prometheus_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid message={incoming_message}, unidentified dialogue.",
        )

    def test_teardown(self):
        """Test the teardown method of the prometheus handler."""
        assert self.prometheus_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a simple behaviour to search a service."""

from typing import cast

from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.simple_service_search.dialogues import OefSearchDialogues
from packages.fetchai.skills.simple_service_search.strategy import Strategy


class ServiceSearchBehaviour(TickerBehaviour):
    """This class provides a simple search behaviour."""

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        pass

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        oef_search_dialogues = cast(
            OefSearchDialogues, self.context.oef_search_dialogues
        )
        search_request, _ = oef_search_dialogues.create(
            counterparty=self.context.search_service_address,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=strategy.get_query(),
        )
        self.context.logger.info("sending search request to OEF search node")
        self.context.outbox.put_message(message=search_request)

    def teardown(self) -> None:
        """
        Implement the task teardown.

        :return: None
        """
        pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the simple_service_registration skill."""

from pathlib import Path

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.models import Description, Location
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.simple_service_registration.strategy import (
    AGENT_LOCATION_MODEL,
    AGENT_REMOVE_SERVICE_MODEL,
    AGENT_SET_SERVICE_MODEL,
    Strategy,
)

from tests.conftest import ROOT_DIR


class TestStrategy(BaseSkillTestCase):
    """Test Strategy of simple_service_registration."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_service_registration"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.location = {"longitude": 0.1270, "latitude": 51.5194}
        cls.service_data = {"key": "seller_service", "value": "generic_service"}

        cls.strategy = Strategy(
            location=cls.location,
            service_data=cls.service_data,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )

    def test__init__(self):
        """Test the __init__ method of the Strategy class."""
        assert self.strategy._remove_service_data == {"key": self.service_data["key"]}

    def test__init__exception(self):
        """Test the __init__ method of the Strategy class where exception is raise."""
        incorrect_service_data_1 = {"key": "seller_service"}
        incorrect_service_data_2 = {
            "incorrect_key": "seller_service",
            "value": "generic_service",
        }
        incorrect_service_data_3 = {
            "key": "seller_service",
            "incorrect_key": "generic_service",
        }
        with pytest.raises(
            AEAEnforceError, match="service_data must contain keys `key` and `value`"
        ):
            Strategy(
                location=self.location, service_data=incorrect_service_data_1,
            )

        with pytest.raises(
            AEAEnforceError, match="service_data must contain keys `key` and `value`"
        ):
            Strategy(
                location=self.location, service_data=incorrect_service_data_2,
            )

        with pytest.raises(
            AEAEnforceError, match="service_data must contain keys `key` and `value`"
        ):
            Strategy(
                location=self.location, service_data=incorrect_service_data_3,
            )

    def test_get_location_description(self):
        """Test the get_location_description method of the Strategy class."""
        description = self.strategy.get_location_description()

        assert type(description) == Description
        assert description.data_model is AGENT_LOCATION_MODEL
        assert description.values.get("location", "") == Location(
            latitude=self.location["latitude"], longitude=self.location["longitude"]
        )

    def test_get_register_service_description(self):
        """Test the get_register_service_description method of the Strategy class."""
        description = self.strategy.get_register_service_description()

        assert type(description) == Description
        assert description.data_model is AGENT_SET_SERVICE_MODEL
        assert description.values.get("key", "") == self.service_data["key"]
        assert description.values.get("value", "") == self.service_data["value"]

    def test_get_unregister_service_description(self):
        """Test the get_unregister_service_description method of the Strategy class."""
        description = self.strategy.get_unregister_service_description()

        assert type(description) == Description
        assert description.data_model is AGENT_REMOVE_SERVICE_MODEL
        assert description.values.get("key", "") == self.service_data["key"]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a scaffold of a handler."""

from typing import Optional

from aea.configurations.base import PublicId
from aea.protocols.base import Message
from aea.skills.base import Handler


class MyScaffoldHandler(Handler):
    """This class scaffolds a handler."""

    SUPPORTED_PROTOCOL = None  # type: Optional[PublicId]

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        raise NotImplementedError

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to an envelope.

        :param message: the message
        :return: None
        """
        raise NotImplementedError

    def teardown(self) -> None:
        """
        Implement the handler teardown.

        :return: None
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the content of quickstart.md file."""
from pathlib import Path

from packages.fetchai.protocols.default.message import DefaultMessage

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import BasePythonMarkdownDocs, extract_code_blocks


class TestQuickstartTest(BasePythonMarkdownDocs):
    """Test the quickstart test."""

    DOC_PATH = Path(ROOT_DIR, "docs", "quickstart.md")


def test_correct_echo_string():
    """Test the echo string in the quickstart is using the correct protocol specification id."""
    file_path = Path(ROOT_DIR, "docs", "quickstart.md")
    bash_code_blocks = extract_code_blocks(filepath=file_path, filter_="bash")
    echo_bloc = bash_code_blocks[19]
    default_protocol_spec_id = echo_bloc.split(",")[2]
    assert (
        str(DefaultMessage.protocol_specification_id) == default_protocol_spec_id
    ), "Spec ids not matching!"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the search feature of the local OEF node."""
import unittest.mock
from typing import cast

import pytest

from aea.common import Address
from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Description,
    Query,
)
from aea.mail.base import Envelope, EnvelopeContext, Message
from aea.multiplexer import InBox, Multiplexer
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.fetchai.connections.local.connection import (
    LocalNode,
    OEFLocalConnection,
    OEF_LOCAL_NODE_ADDRESS,
    OEF_LOCAL_NODE_SEARCH_ADDRESS,
)
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.dialogues import FipaDialogue, FipaDialogues
from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.oef_search.dialogues import OefSearchDialogue
from packages.fetchai.protocols.oef_search.dialogues import (
    OefSearchDialogues as BaseOefSearchDialogues,
)
from packages.fetchai.protocols.oef_search.message import OefSearchMessage

from tests.common.mocks import AnyStringWith
from tests.common.utils import wait_for_condition
from tests.conftest import MAX_FLAKY_RERUNS, _make_local_connection


class OefSearchDialogues(BaseOefSearchDialogues):
    """The dialogues class keeps track of all http dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return OefSearchDialogue.Role.AGENT

        BaseOefSearchDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


class TestNoValidDialogue:
    """Test that the search request returns an empty search result."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.node = LocalNode()
        cls.node.start()

        cls.address_1 = "address_1"
        cls.connection = _make_local_connection(cls.address_1, cls.node,)
        cls.multiplexer = Multiplexer([cls.connection])

        cls.multiplexer.connect()
        cls.dialogues = OefSearchDialogues(cls.address_1)

    @pytest.mark.asyncio
    async def test_wrong_dialogue(self):
        """Test that at the beginning, the search request returns an empty search result."""
        query = Query(
            constraints=[Constraint("foo", ConstraintType("==", 1))], model=None
        )

        # build and send the request
        search_services_request = OefSearchMessage(
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            message_id=2,
            target=1,
            dialogue_reference=self.dialogues.new_self_initiated_dialogue_reference(),
            query=query,
        )
        search_services_request.to = OEF_LOCAL_NODE_SEARCH_ADDRESS

        # the incorrect message cannot be sent into a dialogue, so this is omitted.

        search_services_request.sender = self.address_1
        envelope = Envelope(
            to=search_services_request.to,
            sender=search_services_request.sender,
            message=search_services_request,
        )
        with unittest.mock.patch.object(
            self.node, "_handle_oef_message", side_effect=self.node._handle_oef_message
        ) as mock_handle:
            with unittest.mock.patch.object(self.node.logger, "warning") as mock_logger:
                self.multiplexer.put(envelope)
                wait_for_condition(lambda: mock_handle.called, timeout=1.0)
                mock_logger.assert_any_call(
                    AnyStringWith("Could not create dialogue for message=")
                )

    @classmethod
    def teardown_class(cls):
        """Teardown the test."""
        cls.multiplexer.disconnect()
        cls.node.stop()


class TestEmptySearch:
    """Test that the search request returns an empty search result."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.node = LocalNode()
        cls.node.start()

        cls.address_1 = "address_1"
        cls.multiplexer = Multiplexer(
            [_make_local_connection(cls.address_1, cls.node,)]
        )

        cls.multiplexer.connect()
        cls.dialogues = OefSearchDialogues(cls.address_1)

    def test_empty_search_result(self):
        """Test that at the beginning, the search request returns an empty search result."""
        search_services_request, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=Query(constraints=[], model=None),
        )
        envelope = Envelope(
            to=search_services_request.to,
            sender=search_services_request.sender,
            message=search_services_request,
        )
        self.multiplexer.put(envelope)

        # check the result
        response_envelope = self.multiplexer.get(block=True, timeout=2.0)
        assert (
            response_envelope.protocol_specification_id
            == OefSearchMessage.protocol_specification_id
        )
        search_result = cast(OefSearchMessage, response_envelope.message)
        response_dialogue = self.dialogues.update(search_result)
        assert response_dialogue == sending_dialogue
        assert search_result.performative == OefSearchMessage.Performative.SEARCH_RESULT
        assert search_result.agents == ()

    @classmethod
    def teardown_class(cls):
        """Teardown the test."""
        cls.multiplexer.disconnect()
        cls.node.stop()


class TestSimpleSearchResult:
    """Test that a simple search result return the expected result."""

    def setup(self):
        """Set up the test."""
        self.node = LocalNode()
        self.node.start()

        self.address_1 = "address"
        self.multiplexer = Multiplexer(
            [_make_local_connection(self.address_1, self.node,)]
        )

        self.multiplexer.connect()

        # register a service.
        self.dialogues = OefSearchDialogues(self.address_1)
        self.data_model = DataModel(
            "foobar",
            attributes=[Attribute("foo", int, True), Attribute("bar", str, True)],
        )
        service_description = Description(
            {"foo": 1, "bar": "baz"}, data_model=self.data_model
        )
        register_service_request, self.sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=register_service_request.to,
            sender=register_service_request.sender,
            message=register_service_request,
        )
        self.multiplexer.put(envelope)

    @pytest.mark.flaky(
        reruns=MAX_FLAKY_RERUNS
    )  # TODO: check reasons!. quite unstable test
    def test_not_empty_search_result(self):
        """Test that the search result contains one entry after a successful registration."""
        # build and send the request
        search_services_request, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=Query(constraints=[], model=self.data_model),
        )
        envelope = Envelope(
            to=search_services_request.to,
            sender=search_services_request.sender,
            message=search_services_request,
        )
        self.multiplexer.put(envelope)

        # check the result
        response_envelope = self.multiplexer.get(block=True, timeout=2.0)
        assert (
            response_envelope.protocol_specification_id
            == OefSearchMessage.protocol_specification_id
        )
        search_result = cast(OefSearchMessage, response_envelope.message)
        response_dialogue = self.dialogues.update(search_result)
        assert response_dialogue == sending_dialogue
        assert search_result.performative == OefSearchMessage.Performative.SEARCH_RESULT
        assert search_result.agents == (self.address_1,)

    def teardown(self):
        """Teardown the test."""
        self.multiplexer.disconnect()
        self.node.stop()


class TestUnregister:
    """Test that the unregister service results to Error Message."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.node = LocalNode()
        cls.node.start()

        cls.address_1 = "address_1"
        cls.multiplexer1 = Multiplexer(
            [_make_local_connection(cls.address_1, cls.node,)]
        )
        cls.address_2 = "address_2"
        cls.multiplexer2 = Multiplexer(
            [_make_local_connection(cls.address_2, cls.node,)]
        )
        cls.multiplexer1.connect()
        cls.multiplexer2.connect()
        cls.dialogues = OefSearchDialogues(cls.address_1)

    def test_unregister_service_result(self):
        """Test that at the beginning, the search request returns an empty search result."""
        data_model = DataModel(
            "foobar",
            attributes=[Attribute("foo", int, True), Attribute("bar", str, True)],
        )
        service_description = Description(
            {"foo": 1, "bar": "baz"}, data_model=data_model
        )
        msg, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        self.multiplexer1.put(envelope)

        # check the result
        response_envelope = self.multiplexer1.get(block=True, timeout=5.0)
        assert (
            response_envelope.protocol_specification_id
            == OefSearchMessage.protocol_specification_id
        )
        response = cast(OefSearchMessage, response_envelope.message)
        response_dialogue = self.dialogues.update(response)
        assert response_dialogue == sending_dialogue
        assert response.performative == OefSearchMessage.Performative.OEF_ERROR

        msg, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        self.multiplexer1.put(envelope)

        # Search for the registered service
        msg, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=Query([Constraint("foo", ConstraintType("==", 1))]),
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        self.multiplexer1.put(envelope)
        # check the result
        response_envelope = self.multiplexer1.get(block=True, timeout=5.0)
        search_result = cast(OefSearchMessage, response_envelope.message)
        response_dialogue = self.dialogues.update(search_result)
        assert response_dialogue == sending_dialogue
        assert search_result.performative == OefSearchMessage.Performative.SEARCH_RESULT
        assert len(search_result.agents) == 1

        # unregister the service
        msg, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        self.multiplexer1.put(envelope)

        # the same query returns empty
        # Search for the register agent
        msg, sending_dialogue = self.dialogues.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=Query([Constraint("foo", ConstraintType("==", 1))]),
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        self.multiplexer1.put(envelope)
        # check the result
        response_envelope = self.multiplexer1.get(block=True, timeout=5.0)
        search_result = cast(OefSearchMessage, response_envelope.message)
        response_dialogue = self.dialogues.update(search_result)
        assert response_dialogue == sending_dialogue
        assert search_result.performative == OefSearchMessage.Performative.SEARCH_RESULT
        assert search_result.agents == ()

    @classmethod
    def teardown_class(cls):
        """Teardown the test."""
        cls.multiplexer1.disconnect()
        cls.multiplexer2.disconnect()
        cls.node.stop()


class TestAgentMessage:
    """Test the the OEF will return Dialogue Error if it doesn't know the agent address."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.node = LocalNode()
        cls.node.start()

        cls.address_1 = "address_1"
        cls.multiplexer1 = Multiplexer(
            [_make_local_connection(cls.address_1, cls.node,)]
        )

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return FipaDialogue.Role.SELLER

        cls.dialogues = FipaDialogues(cls.address_1, role_from_first_message)

    @pytest.mark.asyncio
    async def test_messages(self):
        """Test that at the beginning, the search request returns an empty search result."""
        msg, sending_dialogue = self.dialogues.create(
            counterparty="some_agent",
            performative=FipaMessage.Performative.CFP,
            query=Query([Constraint("something", ConstraintType(">", 1))]),
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        with pytest.raises(ConnectionError):
            await _make_local_connection(self.address_1, self.node,).send(envelope)

        self.multiplexer1.connect()
        msg, sending_dialogue = self.dialogues.create(
            counterparty="this_address_does_not_exist",
            performative=FipaMessage.Performative.CFP,
            query=Query([Constraint("something", ConstraintType(">", 1))]),
        )
        envelope = Envelope(to=msg.to, sender=msg.sender, message=msg,)
        self.multiplexer1.put(envelope)

        # check the result
        response_envelope = self.multiplexer1.get(block=True, timeout=5.0)
        assert (
            response_envelope.protocol_specification_id
            == DefaultMessage.protocol_specification_id
        )
        assert response_envelope.sender == OEF_LOCAL_NODE_ADDRESS
        result = response_envelope.message
        assert result.performative == DefaultMessage.Performative.ERROR

    @classmethod
    def teardown_class(cls):
        """Teardown the test."""
        cls.multiplexer1.disconnect()
        cls.node.stop()


class TestFilteredSearchResult:
    """Test that the query system of the search gives the expected result."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.node = LocalNode()
        cls.node.start()

        cls.address_1 = "multiplexer1"
        cls.address_2 = "multiplexer2"
        cls.multiplexer1 = Multiplexer(
            [_make_local_connection(cls.address_1, cls.node,)]
        )
        cls.multiplexer2 = Multiplexer(
            [_make_local_connection(cls.address_2, cls.node,)]
        )
        cls.multiplexer1.connect()
        cls.multiplexer2.connect()
        cls.dialogues1 = OefSearchDialogues(cls.address_1)
        cls.dialogues2 = OefSearchDialogues(cls.address_2)

        # register 'multiplexer1' as a service 'foobar'.
        cls.data_model_foobar = DataModel(
            "foobar",
            attributes=[Attribute("foo", int, True), Attribute("bar", str, True)],
        )
        service_description = Description(
            {"foo": 1, "bar": "baz"}, data_model=cls.data_model_foobar
        )
        register_service_request, sending_dialogue = cls.dialogues1.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=register_service_request.to,
            sender=register_service_request.sender,
            message=register_service_request,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        cls.multiplexer1.put(envelope)
        wait_for_condition(lambda: len(cls.node.services) == 1, timeout=10)

        # register 'multiplexer2' as a service 'barfoo'.
        cls.data_model_barfoo = DataModel(
            "barfoo",
            attributes=[Attribute("foo", int, True), Attribute("bar", str, True)],
        )
        service_description = Description(
            {"foo": 1, "bar": "baz"}, data_model=cls.data_model_barfoo
        )
        register_service_request, sending_dialogue = cls.dialogues2.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=register_service_request.to,
            sender=register_service_request.sender,
            message=register_service_request,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )

        cls.multiplexer2.put(envelope)
        wait_for_condition(lambda: len(cls.node.services) == 2, timeout=10)

        # unregister multiplexer1
        data_model = DataModel(
            "foobar",
            attributes=[Attribute("foo", int, True), Attribute("bar", str, True)],
        )
        service_description = Description(
            {"foo": 1, "bar": "baz"}, data_model=data_model
        )
        msg, sending_dialogue = cls.dialogues1.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            service_description=service_description,
        )
        envelope = Envelope(
            to=msg.to,
            sender=msg.sender,
            message=msg,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        cls.multiplexer1.put(envelope)
        # ensure one service stays registered
        wait_for_condition(lambda: len(cls.node.services) == 1, timeout=10)

    def test_filtered_search_result(self):
        """Test that the search result contains only the entries matching the query."""
        # build and send the request
        search_services_request, sending_dialogue = self.dialogues1.create(
            counterparty=OEF_LOCAL_NODE_SEARCH_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=Query(constraints=[], model=self.data_model_barfoo),
        )
        envelope = Envelope(
            to=search_services_request.to,
            sender=search_services_request.sender,
            message=search_services_request,
            context=EnvelopeContext(connection_id=OEFLocalConnection.connection_id),
        )
        self.multiplexer1.put(envelope)

        # check the result
        response_envelope = InBox(self.multiplexer1).get(block=True, timeout=5.0)
        search_result = cast(OefSearchMessage, response_envelope.message)
        response_dialogue = self.dialogues1.update(search_result)
        assert response_dialogue == sending_dialogue
        assert search_result.performative == OefSearchMessage.Performative.SEARCH_RESULT
        assert search_result.agents == (self.address_2,), self.node.services

    @classmethod
    def teardown_class(cls):
        """Teardown the test."""
        cls.multiplexer1.disconnect()
        cls.multiplexer2.disconnect()
        cls.node.stop()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Core definitions for the AEA command-line tool."""
from typing import Optional

import click

import aea
from aea.cli.add import add
from aea.cli.add_key import add_key
from aea.cli.build import build
from aea.cli.config import config
from aea.cli.create import create
from aea.cli.delete import delete
from aea.cli.eject import eject
from aea.cli.fetch import fetch
from aea.cli.fingerprint import fingerprint
from aea.cli.freeze import freeze
from aea.cli.generate import generate
from aea.cli.generate_key import generate_key
from aea.cli.generate_wealth import generate_wealth
from aea.cli.get_address import get_address
from aea.cli.get_multiaddress import get_multiaddress
from aea.cli.get_wealth import get_wealth
from aea.cli.init import init
from aea.cli.install import install
from aea.cli.interact import interact
from aea.cli.issue_certificates import issue_certificates
from aea.cli.launch import launch
from aea.cli.list import list_command as _list
from aea.cli.local_registry_sync import local_registry_sync
from aea.cli.login import login
from aea.cli.logout import logout
from aea.cli.publish import publish
from aea.cli.push import push
from aea.cli.register import register
from aea.cli.remove import remove
from aea.cli.remove_key import remove_key
from aea.cli.reset_password import reset_password
from aea.cli.run import run
from aea.cli.scaffold import scaffold
from aea.cli.search import search
from aea.cli.transfer import transfer
from aea.cli.upgrade import upgrade
from aea.cli.utils.click_utils import registry_path_option
from aea.cli.utils.context import Context
from aea.cli.utils.loggers import logger, simple_verbosity_option
from aea.helpers.win32 import enable_ctrl_c_support


@click.group(name="aea")  # type: ignore
@click.version_option(aea.__version__, prog_name="aea")
@simple_verbosity_option(logger, default="INFO")
@click.option(
    "-s",
    "--skip-consistency-check",
    "skip_consistency_check",
    is_flag=True,
    required=False,
    default=False,
    help="Skip consistency checks of agent during command execution.",
)
@registry_path_option
@click.pass_context
def cli(
    click_context: click.Context,
    skip_consistency_check: bool,
    registry_path: Optional[str],
) -> None:
    """Command-line tool for setting up an Autonomous Economic Agent (AEA)."""
    verbosity_option = click_context.meta.pop("verbosity")
    click_context.obj = Context(cwd=".", verbosity=verbosity_option)
    click_context.obj.registry_path = registry_path
    click_context.obj.set_config("skip_consistency_check", skip_consistency_check)

    # enables CTRL+C support on windows!
    enable_ctrl_c_support()


cli.add_command(_list)
cli.add_command(add_key)
cli.add_command(add)
cli.add_command(build)
cli.add_command(create)
cli.add_command(config)
cli.add_command(delete)
cli.add_command(eject)
cli.add_command(fetch)
cli.add_command(fingerprint)
cli.add_command(freeze)
cli.add_command(generate_key)
cli.add_command(generate_wealth)
cli.add_command(generate)
cli.add_command(get_address)
cli.add_command(get_multiaddress)
cli.add_command(get_wealth)
cli.add_command(init)
cli.add_command(install)
cli.add_command(interact)
cli.add_command(issue_certificates)
cli.add_command(launch)
cli.add_command(login)
cli.add_command(logout)
cli.add_command(publish)
cli.add_command(push)
cli.add_command(register)
cli.add_command(remove)
cli.add_command(remove_key)
cli.add_command(reset_password)
cli.add_command(run)
cli.add_command(scaffold)
cli.add_command(search)
cli.add_command(local_registry_sync)
cli.add_command(transfer)
cli.add_command(upgrade)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the class to connect to an oracle client contract."""

import logging

from aea_ledger_ethereum import EthereumApi

from aea.common import Address
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi


PUBLIC_ID = PublicId.from_str("fetchai/oracle_client:0.4.0")

_default_logger = logging.getLogger(
    "aea.packages.fetchai.contracts.oracle_client.contract"
)


class FetchOracleClientContract(Contract):
    """The Fetch oracle client contract."""

    @classmethod
    def get_query_transaction(
        cls,
        ledger_api: LedgerApi,
        contract_address: Address,
        from_address: Address,
        query_function: str,
        gas: int = 0,
    ) -> None:
        """
        Get transaction to query oracle value in contract

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param from_address: the address of the transaction sender.
        :param query_function: the query oracle value function.
        :param gas: the gas limit for the transaction.
        :return: None
        """
        if ledger_api.identifier == EthereumApi.identifier:
            nonce = ledger_api.api.eth.getTransactionCount(from_address)
            instance = cls.get_instance(ledger_api, contract_address)
            function = getattr(instance.functions, query_function)
            query_args = ()
            intermediate = function(*query_args)
            tx = intermediate.buildTransaction(
                {
                    "gas": gas,
                    "gasPrice": ledger_api.api.toWei("50", "gwei"),
                    "nonce": nonce,
                }
            )
            tx = ledger_api.update_with_gas_estimate(tx)
            return tx
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the implementation of a 'dummy' connection useful for testing."""

import asyncio
from concurrent.futures._base import CancelledError
from typing import Optional

from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.mail.base import Envelope


class DummyConnection(Connection):
    """A dummy connection that just stores the messages."""

    connection_id = PublicId.from_str("fetchai/dummy:0.1.0")

    def __init__(self, **kwargs):
        """Initialize."""
        super().__init__(**kwargs)
        self.state = ConnectionStates.disconnected
        self._queue = None

    async def connect(self, *args, **kwargs):
        """Connect."""
        self._queue = asyncio.Queue(loop=self.loop)
        self.state = ConnectionStates.connected

    async def disconnect(self, *args, **kwargs):
        """Disconnect."""
        assert self._queue is not None
        await self._queue.put(None)
        self.state = ConnectionStates.disconnected

    async def send(self, envelope: "Envelope"):
        """Send an envelope."""
        assert self._queue is not None
        self._queue.put_nowait(envelope)

    async def receive(self, *args, **kwargs) -> Optional["Envelope"]:
        """Receive an envelope."""
        try:
            assert self._queue is not None
            envelope = await self._queue.get()
            if envelope is None:
                return None
            return envelope
        except CancelledError:
            return None
        except Exception as e:
            print(str(e))
            return None

    def put(self, envelope: Envelope):
        """Put an envelope in the queue."""
        assert self._queue is not None
        self._queue.put_nowait(envelope)
